#
# Copyright (C) 2013-2024 The ESPResSo project
#
# This file is part of ESPResSo.
#
# ESPResSo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ESPResSo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import espressomd
import espressomd.version
import espressomd.electrostatics
import argparse
import time
import numpy as np

parser = argparse.ArgumentParser(description="Benchmark P3M simulations.")
parser.add_argument("--size", metavar="S", action="store",
                    default=9, required=False, type=int,
                    help="Problem size, such that the number of particles N is "
                         "equal to (2*S)^2; with --weak-scaling this number N "
                         "is multiplied by the number of cores!")
parser.add_argument("--gpu", action=argparse.BooleanOptionalAction,
                    default=False, required=False, help="Use GPU implementation")
parser.add_argument("--topology", metavar=("X", "Y", "Z"), nargs=3, action="store",
                    default=None, required=False, type=int, help="Cartesian topology")
group = parser.add_mutually_exclusive_group()
group.add_argument("--weak-scaling", action="store_true",
                   help="Weak scaling benchmark (Gustafson's law: constant work per core)")
group.add_argument("--strong-scaling", action="store_true",
                   help="Strong scaling benchmark (Amdahl's law: constant total work)")
args = parser.parse_args()


def get_reference_values_per_ion(base_vector):
    madelung_constant = -1.74756459463318219
    base_tensor = base_vector * np.eye(3)
    ref_energy = madelung_constant
    ref_pressure = madelung_constant * base_tensor / np.trace(base_tensor)
    return ref_energy, ref_pressure


def get_normalized_values_per_ion(system):
    energy = system.analysis.energy()["coulomb"]
    p_scalar = system.analysis.pressure()["coulomb"]
    p_tensor = system.analysis.pressure_tensor()["coulomb"]
    N = len(system.part)
    V = system.volume()
    return 2. * energy / N, 2. * p_scalar * V / N, 2. * p_tensor * V / N


# initialize system
system = espressomd.System(box_l=[100., 100., 100.])
system.time_step = 0.01
system.cell_system.skin = 0.4

# set MPI Cartesian topology
node_grid = system.cell_system.node_grid.copy()
n_cores = int(np.prod(node_grid))
if args.topology:
    system.cell_system.node_grid = node_grid = args.topology

# place ions on a cubic lattice
base_vector = np.array([1., 1., 1.])
lattice_size = 3 * [2 * args.size]
if args.weak_scaling:
    lattice_size = np.multiply(lattice_size, node_grid)
system.box_l = np.multiply(lattice_size, base_vector)
for var_j in range(lattice_size[0]):
    for var_k in range(lattice_size[1]):
        for var_l in range(lattice_size[2]):
            _ = system.part.add(pos=np.multiply([var_j, var_k, var_l], base_vector),
                                q=(-1.)**(var_j + var_k + var_l), fix=3 * [True])

# setup P3M algorithm
algorithm = espressomd.electrostatics.P3M
if args.gpu:
    algorithm = espressomd.electrostatics.P3MGPU
solver = algorithm(prefactor=1., accuracy=1e-6)
if (espressomd.version.major(), espressomd.version.minor()) == (4, 2):
    system.actors.add(solver)
else:
    system.electrostatics.solver = solver


print("Algorithm executed. \n")

# Old rtol_pressure = 2e-5
# This resulted in failures especially at high number of nodes therefore increased
# to a larger value.

atol_energy = atol_pressure = 1e-12
atol_forces = 1e-5
atol_abs_forces = 2e-6

rtol_energy = 5e-6
rtol_pressure = 1e-4
rtol_forces = 0.
rtol_abs_forces = 0.
# run checks
print("Executing sanity checks...\n")
forces = np.copy(system.part.all().f)
energy, p_scalar, p_tensor = get_normalized_values_per_ion(system)
ref_energy, ref_pressure = get_reference_values_per_ion(base_vector)
np.testing.assert_allclose(energy, ref_energy, atol=atol_energy, rtol=rtol_energy)
np.testing.assert_allclose(p_scalar, np.trace(ref_pressure) / 3.,
                           atol=atol_pressure, rtol=rtol_pressure)
np.testing.assert_allclose(p_tensor, ref_pressure, atol=atol_pressure, rtol=rtol_pressure)
np.testing.assert_allclose(forces, 0., atol=atol_forces, rtol=rtol_forces)
np.testing.assert_allclose(np.median(np.abs(forces)), 0., atol=atol_abs_forces, rtol=rtol_abs_forces)

print("Final convergence met with tolerances: \n\
            energy: ", atol_energy, "\n\
            p_scalar: ", atol_pressure, "\n\
            p_tensor: ", atol_pressure, "\n\
            forces: ", atol_forces, "\n\
            abs_forces: ", atol_abs_forces, "\n")

print("Sampling runtime...\n")
# sample runtime
n_steps = 10
timings = []
for _ in range(10):
    tick = time.time()
    system.integrator.run(n_steps)
    tock = time.time()
    timings.append((tock - tick) / n_steps)

print("10 steps executed...\n")
# write results to file
header = '"mode","cores","mpi.x","mpi.y","mpi.z","particles","mean","std"\n'
report = f'''"{"weak scaling" if args.weak_scaling else "strong scaling"}",\
{n_cores},{node_grid[0]},{node_grid[1]},{node_grid[2]},{len(system.part)},\
{np.mean(timings):.3e},{np.std(timings,ddof=1):.3e}\n'''
print(header)
print(report)

print(f"Performance: {np.mean(timings):.3e} \n")



import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import *  # noqa: F403

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# This is an example configuration file
site_configuration = {
    'systems': [
        {
            'name': 'deucalion',
            'descr': 'Deucalion, a EuroHPC JU system',
            'modules_system': 'lmod',
            'hostnames': ['ln*', 'cn*', 'gn*'],
            'prefix': reframe_prefix,
            'partitions': [
                {
                    'name': 'arm',
                    'scheduler': 'slurm',
                    'prepare_cmds': [
                        'wrap.sh << EOF',
                        # bypass CPU autodetection for now aarch64/a64fx,
                        # see https://github.com/EESSI/software-layer/pull/608
                        'export EESSI_SOFTWARE_SUBDIR_OVERRIDE=aarch64/a64fx',
                        common_eessi_init(),
                        # Pass job environment variables like $PATH, etc., into job steps
                        'export SLURM_EXPORT_ENV=HOME,PATH,LD_LIBRARY_PATH,PYTHONPATH',
                    ],
                    'launcher': 'mpirun',
                    # Use --export=None to avoid that login environment is passed down to submitted jobs
                    'access': ['-p normal-arm', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 120,
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        # NB: Deucalion's MaxMemPerNode is undefined. Experimentally I found you cannot submit with
                        # more than --mem=30000M
                        'mem_per_node': 30000  # in MiB
                    },
                    'descr': 'CPU ARM A64FX partition, see https://docs.macc.fccn.pt/deucalion/#compute-nodes'
                },
            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}



"""
This module tests the binary 'osu' in available modules containing substring 'OSU-Micro-Benchmarks'. The basic
application class is taken from the hpctestlib to which extra features are added.

Note: OSU-Micro-Benchmarks CUDA module binaries must be linked to stubs so that it at the least finds libcuda.so.1 on
non-GPU nodes. Otherwise those tests will FAIL.
"""
import reframe as rfm
from reframe.core.builtins import parameter, run_after
from reframe.utility import reframe

from hpctestlib.microbenchmarks.mpi.osu import osu_benchmark

from eessi.testsuite.constants import COMPUTE_UNIT, CPU, DEVICE_TYPES, INVALID_SYSTEM, GPU, NODE, SCALES
from eessi.testsuite.eessi_mixin import EESSI_Mixin
from eessi.testsuite.utils import find_modules, log


def filter_scales_pt2pt_cpu():
    """
    Filtering function for filtering scales for the pt2pt OSU test on CPUs
    returns all scales with either 2 cores, 1 full node, or 2 full nodes
    """
    return [
        k for (k, v) in SCALES.items()
        if v['num_nodes'] * v.get('num_cpus_per_node', 0) == 2
        or (v['num_nodes'] == 2 and v.get('node_part', 0) == 1)
        or (v['num_nodes'] == 1 and v.get('node_part', 0) == 1)
    ]


def filter_scales_pt2pt_gpu():
    """
    Filtering function for filtering scales for the pt2pt OSU test on GPUs
    returns all scales with either a partial node, 1 full node, or 2 full nodes
    """
    return [
        k for (k, v) in SCALES.items()
        if (v['num_nodes'] == 1 and v.get('node_part', 0) > 1)
        or (v['num_nodes'] == 2 and v.get('node_part', 0) == 1)
        or (v['num_nodes'] == 1 and v.get('node_part', 0) == 1)
    ]


def filter_scales_coll():
    """
    Filtering function for filtering scales for the collective OSU test
    returns all scales with at least 2 cores
    """
    return [
        k for (k, v) in SCALES.items()
        if (v['num_nodes'] * v.get('num_cpus_per_node', 1) > 1)
        or (v.get('node_part', 0) > 0)
    ]


class EESSI_OSU_Base(osu_benchmark):
    """ base class for OSU tests """
    time_limit = '30m'
    module_name = parameter(find_modules('OSU-Micro-Benchmarks'))

    # reset num_tasks_per_node from the hpctestlib: we handle it ourselves
    num_tasks_per_node = None
    readonly_files = ['']

    # Set num_warmup_iters to 5 to reduce execution time, especially on slower interconnects
    num_warmup_iters = 5
    # Set num_iters to 10 to reduce execution time, especially on slower interconnects
    num_iters = 10

    def required_mem_per_node(self):
        return self.num_tasks_per_node * 1024

    @run_after('init')
    def filter_scales_2gpus(self):
        """Filter out scales with < 2 GPUs if running on GPUs"""
        if (
            self.device_type == DEVICE_TYPES[GPU]
            and SCALES[self.scale]['num_nodes'] == 1
            and SCALES[self.scale].get('num_gpus_per_node', 2) < 2
        ):
            self.valid_systems = [INVALID_SYSTEM]
            log(f'valid_systems set to {self.valid_systems} for scale {self.scale} and device_type {self.device_type}')

    @run_after('init')
    def set_device_buffers(self):
        """
        device_buffers is inherited from the hpctestlib class and adds options to the launcher
        commands in a @run_before('setup') hook if not equal to 'cpu'.
        Therefore, we must set device_buffers *before* the @run_before('setup') hooks.
        """
        if self.device_type == DEVICE_TYPES[GPU]:
            self.device_buffers = 'cuda'
        else:
            self.device_buffers = 'cpu'

    @run_after('init')
    def set_tags(self):
        """ Setting custom tags """
        self.bench_name = self.benchmark_info[0]
        self.tags.add(self.bench_name.split('.')[-1])


class EESSI_OSU_pt2pt_Base(EESSI_OSU_Base):
    ''' point-to-point OSU test base class '''
    compute_unit = COMPUTE_UNIT[NODE]

    @run_after('init')
    def filter_benchmark_pt2pt(self):
        """ Filter out all non-mpi.pt2pt benchmarks """
        if not self.benchmark_info[0].startswith('mpi.pt2pt'):
            self.valid_systems = [INVALID_SYSTEM]

    @run_after('init')
    def select_ci(self):
        " Select the CI variants "
        if (self.bench_name in ['mpi.pt2pt.osu_latency', 'mpi.pt2pt.osu_bw']):
            self.bench_name_ci = self.bench_name

    @run_after('init')
    def set_num_tasks_per_compute_unit(self):
        """ Setting number of tasks per compute unit and cpus per task. This sets num_cpus_per_task
        for 1 node and 2 node options where the request is for full nodes."""
        if SCALES.get(self.scale).get('num_nodes') == 1:
            self.num_tasks_per_compute_unit = 2

    @run_after('setup')
    def adjust_executable_opts(self):
        """The option "D D" is only meant for Devices if and not for CPU tests.
        This option is added by hpctestlib in a @run_before('setup') to all pt2pt tests which is not required.
        Therefore we must override it *after* the 'setup' phase
        """
        if self.device_type == DEVICE_TYPES[CPU]:
            self.executable_opts = [x for x in self.executable_opts if x != 'D']


@rfm.simple_test
class EESSI_OSU_pt2pt_CPU(EESSI_OSU_pt2pt_Base, EESSI_Mixin):
    ''' point-to-point OSU test on CPUs'''
    scale = parameter(filter_scales_pt2pt_cpu())
    device_type = DEVICE_TYPES[CPU]


@rfm.simple_test
class EESSI_OSU_pt2pt_GPU(EESSI_OSU_pt2pt_Base, EESSI_Mixin):
    ''' point-to-point OSU test on GPUs'''
    scale = parameter(filter_scales_pt2pt_gpu())
    device_type = DEVICE_TYPES[GPU]
    always_request_gpus = True

    @run_after('setup')
    def skip_test_gpus(self):
        num_gpus = self.num_gpus_per_node * self.num_nodes
        if self.scale not in ['1_node', '2_nodes']:
            # On a partial node allocation, run this test only if exactly 2 GPUs are allocated
            self.skip_if(
                num_gpus != 2,
                f"Skipping test : {num_gpus} GPU(s) available for this test case, need exactly 2"
            )
        elif self.scale == '1_node':
            # Make sure there are at least 2 GPUs
            self.skip_if(num_gpus < 2, "Skipping GPU test : only 1 GPU available for this test case")


@rfm.simple_test
class EESSI_OSU_coll(EESSI_OSU_Base, EESSI_Mixin):
    ''' collective OSU test '''
    scale = parameter(filter_scales_coll())
    device_type = parameter([DEVICE_TYPES[CPU], DEVICE_TYPES[GPU]])

    @run_after('init')
    def filter_benchmark_coll(self):
        """ Filter out all non-mpi.collective benchmarks """
        if not self.benchmark_info[0].startswith('mpi.collective'):
            self.valid_systems = [INVALID_SYSTEM]

    @run_after('init')
    def select_ci(self):
        " Select the CI variants "
        if (self.bench_name in ['mpi.collective.osu_allreduce', 'mpi.collective.osu_alltoall']):
            self.bench_name_ci = self.bench_name

    @run_after('init')
    def set_compute_unit(self):
        """
        Set the compute unit to which tasks will be assigned:
        one task per core for CPU runs, and one task per GPU for GPU runs.
        """
        device_to_compute_unit = {
            DEVICE_TYPES[CPU]: COMPUTE_UNIT[CPU],
            DEVICE_TYPES[GPU]: COMPUTE_UNIT[GPU],
        }
        self.compute_unit = device_to_compute_unit.get(self.device_type)

    @run_after('setup')
    def skip_test_1gpu(self):
        if self.device_type == DEVICE_TYPES[GPU]:
            num_gpus = self.num_gpus_per_node * self.num_nodes
            self.skip_if(num_gpus < 2, "Skipping GPU test : only 1 GPU available for this test case")



"""
This module tests the binary 'lmp' in available modules containing substring 'LAMMPS'.
The tests come from the lammps github repository (https://github.com/lammps/lammps/)
"""

import reframe as rfm
from reframe.core.builtins import deferrable, parameter, performance_function, run_after, sanity_function
import reframe.utility.sanity as sn

from eessi.testsuite import utils
from eessi.testsuite.constants import CI, COMPUTE_UNIT, CPU, DEVICE_TYPES, GPU, TAGS
from eessi.testsuite.eessi_mixin import EESSI_Mixin


class EESSI_LAMMPS_base(rfm.RunOnlyRegressionTest):
    time_limit = '30m'
    device_type = parameter([DEVICE_TYPES[CPU], DEVICE_TYPES[GPU]])

    # Parameterize over all modules that start with LAMMPS
    module_name = parameter(utils.find_modules('LAMMPS'))

    def required_mem_per_node(self):
        mem = {'slope': 0.07, 'intercept': 0.5}
        return (self.num_tasks_per_node * mem['slope'] + mem['intercept']) * 1024

    # Set sanity step
    @deferrable
    def assert_lammps_openmp_treads(self):
        '''Assert that OpenMP thread(s) per MPI task is set'''
        n_threads = sn.extractsingle(
            r'^  using (?P<threads>[0-9]+) OpenMP thread\(s\) per MPI task', self.stdout, 'threads', int)
        utils.log(f'OpenMP thread(s) is {n_threads}')

        return sn.assert_eq(n_threads, self.num_cpus_per_task)

    @deferrable
    def assert_lammps_processor_grid(self):
        '''Assert that the processor grid is set correctly'''
        grid = list(sn.extractall(
            '^  (?P<x>[0-9]+) by (?P<y>[0-9]+) by (?P<z>[0-9]+) MPI processor grid', self.stdout, tag=['x', 'y', 'z']))
        n_cpus = int(grid[0][0]) * int(grid[0][1]) * int(grid[0][2])

        return sn.assert_eq(n_cpus, self.num_tasks)

    @deferrable
    def assert_run(self):
        '''Assert that the test calulated the right number of neighbours'''
        regex = r'^Loop time of (?P<perf>[.0-9]+) on [0-9]+ procs for 100 steps with (?P<atoms>\S+) atoms'
        n_atoms = sn.extractsingle(regex, self.stdout, 'atoms', int)

        return sn.assert_eq(n_atoms, 32000)

    @run_after('init')
    def set_compute_unit(self):
        """Set the compute unit to which tasks will be assigned """
        if self.device_type == 'cpu':
            self.compute_unit = COMPUTE_UNIT['CPU']
        elif self.device_type == 'gpu':
            self.compute_unit = COMPUTE_UNIT['GPU']
        else:
            msg = f"No mapping of device type {self.device_type} to a COMPUTE_UNIT was specified in this test"
            raise NotImplementedError(msg)


@rfm.simple_test
class EESSI_LAMMPS_lj(EESSI_LAMMPS_base, EESSI_Mixin):
    tags = {TAGS[CI]}

    sourcesdir = 'src/lj'
    readonly_files = ['in.lj']
    executable = 'lmp -in in.lj'

    @deferrable
    def check_number_neighbors(self):
        '''Assert that the test calulated the right number of neighbours'''
        regex = r'Neighbor list builds = (?P<neigh>\S+)'
        n_neigh = sn.extractsingle(regex, self.stdout, 'neigh', int)
        return sn.assert_eq(n_neigh, 5)

    @deferrable
    def assert_energy(self):
        '''Asert that the calculated energy at timestep 100 is with the margin of error'''
        regex = r'^\s+100\s+[-+]?[.0-9]+\s+[-+]?[.0-9]+\s+0\s+(?P<energy>[-+]?[.0-9]+)'
        energy = sn.extractsingle(regex, self.stdout, 'energy', float)
        energy_diff = sn.abs(energy - (-4.6223613))
        return sn.assert_lt(energy_diff, 1e-4)

    @performance_function('timesteps/s')
    def perf(self):
        regex = r'^Performance: [.0-9]+ tau/day, (?P<perf>[.0-9]+) timesteps/s'
        return sn.extractsingle(regex, self.stdout, 'perf', float)

    @sanity_function
    def assert_sanity(self):
        '''Check all sanity criteria'''
        return sn.all([
            self.assert_lammps_openmp_treads(),
            self.assert_lammps_processor_grid(),
            self.assert_run(),
            self.check_number_neighbors(),
            self.assert_energy(),
        ])

    @run_after('setup')
    def set_executable_opts(self):
        """Set executable opts based on device_type parameter"""
        # should also check if LAMMPS is installed with kokkos.
        # Because this executable opt is only for that case.
        if self.device_type == "gpu":
            if 'kokkos' in self.module_name:
                self.executable_opts += [
                    f'-kokkos on t {self.num_cpus_per_task} g {self.num_gpus_per_node}',
                    '-suffix kk',
                    '-package kokkos newton on neigh half',
                ]
                utils.log(f'executable_opts set to {self.executable_opts}')
            else:
                self.executable_opts += [f'-suffix gpu -package gpu {self.num_gpus_per_node}']
                utils.log(f'executable_opts set to {self.executable_opts}')


@rfm.simple_test
class EESSI_LAMMPS_rhodo(EESSI_LAMMPS_base, EESSI_Mixin):
    sourcesdir = 'src/rhodo'
    executable = 'lmp -in in.rhodo'
    readonly_files = ['data.rhodo', 'in.rhodo']

    @deferrable
    def check_number_neighbors(self):
        '''Assert that the test calulated the right number of neighbours'''
        regex = r'Neighbor list builds = (?P<neigh>\S+)'
        n_neigh = sn.extractsingle(regex, self.stdout, 'neigh', int)
        return sn.assert_eq(n_neigh, 11)

    @deferrable
    def assert_energy(self):
        '''Asert that the calculated energy at timestep 100 is with the margin of error'''
        regex = r'^-+\s+Step\s+100\s+-+\s+CPU\s=\s+[.0-9]+\s+\(sec\)\s+-+\nTotEng\s+=\s+(?P<energy>[-+]?[.0-9]+)'
        energy = sn.extractsingle(regex, self.stdout, 'energy', float)
        energy_diff = sn.abs(energy - (-25290.7300))
        return sn.assert_lt(energy_diff, 1e-1)

    @performance_function('timesteps/s')
    def perf(self):
        regex = r'^Performance: [.0-9]+ ns/day, [.0-9]+ hours/ns, (?P<perf>[.0-9]+) timesteps/s'
        return sn.extractsingle(regex, self.stdout, 'perf', float)

    @sanity_function
    def assert_sanity(self):
        '''Check all sanity criteria'''
        return sn.all([
            self.assert_lammps_openmp_treads(),
            self.assert_lammps_processor_grid(),
            self.assert_run(),
            self.check_number_neighbors(),
            self.assert_energy(),
        ])

    @run_after('setup')
    def set_executable_opts(self):
        """Set executable opts based on device_type parameter"""
        # should also check if the lammps is installed with kokkos.
        # Because this executable opt is only for that case.
        if self.device_type == "gpu":
            if 'kokkos' in self.module_name:
                self.executable_opts += [
                    f'-kokkos on t {self.num_cpus_per_task} g {self.num_gpus_per_node}',
                    '-suffix kk',
                    '-package kokkos newton on neigh half',
                ]
                utils.log(f'executable_opts set to {self.executable_opts}')
            else:
                self.executable_opts += [f'-suffix gpu -package gpu {self.num_gpus_per_node}']
                utils.log(f'executable_opts set to {self.executable_opts}')



#
# Copyright (C) 2018-2024 The ESPResSo project
#
# This file is part of ESPResSo.
#
# ESPResSo is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ESPResSo is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import time
import espressomd
import numpy as np

required_features = ["LENNARD_JONES"]
espressomd.assert_features(required_features)

parser = argparse.ArgumentParser(description="Benchmark LJ simulations.")
parser.add_argument("--particles-per-core", metavar="N", action="store",
                    type=int, default=2000, required=False,
                    help="Number of particles in the simulation box")
parser.add_argument("--sample-size", metavar="S", action="store",
                    type=int, default=30, required=False,
                    help="Sample size")
parser.add_argument("--volume-fraction", metavar="FRAC", action="store",
                    type=float, default=0.50, required=False,
                    help="Fraction of the simulation box volume occupied by "
                    "particles (range: [0.01-0.74], default: 0.50)")
args = parser.parse_args()

# process and check arguments
measurement_steps = 100
if args.particles_per_core < 16000:
    measurement_steps = 200
if args.particles_per_core < 10000:
    measurement_steps = 500
if args.particles_per_core < 5000:
    measurement_steps = 1000
if args.particles_per_core < 1000:
    measurement_steps = 2000
if args.particles_per_core < 600:
    measurement_steps = 4000
if args.particles_per_core < 260:
    measurement_steps = 6000
assert args.volume_fraction > 0., "volume_fraction must be a positive number"
assert args.volume_fraction < np.pi / (3. * np.sqrt(2.)), \
    "volume_fraction exceeds the physical limit of sphere packing (~0.74)"

# make simulation deterministic
np.random.seed(42)


def get_reference_values_per_atom(x):
    # result of a polynomial fit in the range from 0.01 to 0.55
    energy = 54.2 * x**3 - 23.8 * x**2 + 4.6 * x - 0.09
    pressure = 377. * x**3 - 149. * x**2 + 32.2 * x - 0.58
    return energy, pressure


def get_normalized_values_per_atom(system):
    energy = system.analysis.energy()["non_bonded"]
    pressure = system.analysis.pressure()["non_bonded"]
    N = len(system.part)
    V = system.volume()
    return 2. * energy / N, 2. * pressure * V / N


system = espressomd.System(box_l=[10., 10., 10.])
system.time_step = 0.01
system.cell_system.skin = 0.5

lj_eps = 1.0  # LJ epsilon
lj_sig = 1.0  # particle diameter
lj_cut = lj_sig * 2**(1. / 6.)  # cutoff distance

n_proc = system.cell_system.get_state()["n_nodes"]
n_part = n_proc * args.particles_per_core
node_grid = np.array(system.cell_system.node_grid)
# volume of N spheres with radius r: N * (4/3*pi*r^3)
box_v = args.particles_per_core * 4. / 3. * \
    np.pi * (lj_sig / 2.)**3 / args.volume_fraction
# box_v = (x * n) * x * x for a column
system.box_l = float((box_v)**(1. / 3.)) * node_grid
assert np.abs(n_part * 4. / 3. * np.pi * (lj_sig / 2.)**3 / np.prod(system.box_l) - args.volume_fraction) < 0.1

system.non_bonded_inter[0, 0].lennard_jones.set_params(
    epsilon=lj_eps, sigma=lj_sig, cutoff=lj_cut, shift="auto")

system.part.add(pos=np.random.random((n_part, 3)) * system.box_l)

# energy minimization
max_steps = 1000
# particle forces for volume fractions between 0.1 and 0.5 follow a polynomial
target_f_max = 20. * args.volume_fraction**2
system.integrator.set_steepest_descent(
    f_max=target_f_max, gamma=0.001, max_displacement=0.01 * lj_sig)
n_steps = system.integrator.run(max_steps)
assert n_steps < max_steps, f'''energy minimization failed: \
E = {system.analysis.energy()["total"] / len(system.part):.3g} per particle, \
f_max = {np.max(np.linalg.norm(system.part.all().f, axis=1)):.2g}, \
target f_max = {target_f_max:.2g}'''

# warmup
system.integrator.set_vv()
system.thermostat.set_langevin(kT=1.0, gamma=1.0, seed=42)

# tuning and equilibration
min_skin = 0.2
max_skin = 1.0
print("Tune skin: {:.3f}".format(system.cell_system.tune_skin(
    min_skin=min_skin, max_skin=max_skin, tol=0.05, int_steps=100)))
print("Equilibration")
system.integrator.run(min(5 * measurement_steps, 60000))
print("Tune skin: {:.3f}".format(system.cell_system.tune_skin(
    min_skin=min_skin, max_skin=max_skin, tol=0.05, int_steps=100)))
print("Equilibration")
system.integrator.run(min(10 * measurement_steps, 60000))

print("Sampling runtime...")
timings = []
energies = []
pressures = []
for i in range(args.sample_size):
    tick = time.time()
    system.integrator.run(measurement_steps)
    tock = time.time()
    t = (tock - tick) / measurement_steps
    timings.append(t)
    energy, pressure = get_normalized_values_per_atom(system)
    energies.append(energy)
    pressures.append(pressure)

sim_energy = np.mean(energies)
sim_pressure = np.mean(pressures)
ref_energy, ref_pressure = get_reference_values_per_atom(args.volume_fraction)

print("Algorithm executed. \n")
np.testing.assert_allclose(sim_energy, ref_energy, atol=0., rtol=0.1)
np.testing.assert_allclose(sim_pressure, ref_pressure, atol=0., rtol=0.1)

print("Final convergence met with relative tolerances: \n\
            sim_energy: ", 0.1, "\n\
            sim_pressure: ", 0.1, "\n")

header = '"mode","cores","mpi.x","mpi.y","mpi.z","particles","volume_fraction","mean","std"'
report = f'''"weak scaling",{n_proc},{node_grid[0]},{node_grid[1]},\
{node_grid[2]},{len(system.part)},{args.volume_fraction:.4f},\
{np.mean(timings):.3e},{np.std(timings,ddof=1):.3e}'''
print(header)
print(report)
print(f"Performance: {np.mean(timings):.3e}")



"""
This module tests mpi4py's MPI_Reduce call
"""

import reframe as rfm
import reframe.utility.sanity as sn

# added only to make the linter happy
from reframe.core.builtins import variable, parameter, run_after, performance_function, sanity_function

from eessi.testsuite import hooks
from eessi.testsuite.constants import SCALES, COMPUTE_UNIT, CPU
from eessi.testsuite.utils import find_modules


# This python decorator indicates to ReFrame that this class defines a test
# Our class inherits from rfm.RunOnlyRegressionTest, since this test does not have a compilation stage
# https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RunOnlyRegressionTest
@rfm.simple_test
class EESSI_MPI4PY(rfm.RunOnlyRegressionTest):
    # Programming environments are only relevant for tests that compile something
    # Since we are testing existing modules, we typically don't compile anything and simply define
    # 'default' as the valid programming environment
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.valid_prog_environs
    valid_prog_environs = ['default']

    # Typically, we list here the name of our cluster as it is specified in our ReFrame configuration file
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.valid_systems
    valid_systems = ['*']

    # ReFrame will generate a test for each module
    # NOTE: each parameter adds a new dimension to the parametrization space.
    # (EG 4 parameters with (3,3,2,2) possible values will result in 36 tests).
    # Be mindful of how many parameters you add to avoid the number of tests generated being excessive.
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.parameter
    module_name = parameter(find_modules('mpi4py'))

    # ReFrame will generate a test for each scale
    scale = parameter(SCALES.keys())

    # Our script has two arguments, --n_iter and --n_warmup. By defining these as ReFrame variables, we can
    # enable the end-user to overwrite their value on the command line when invoking ReFrame.
    # Note that we don't typically expose ALL variables, especially if a script has many - we expose
    # only those that we think an end-user might want to overwrite
    # Number of iterations to run (more iterations takes longer, but results in more accurate timing)
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.variable
    n_iterations = variable(int, value=1000)

    # Similar for the number of warmup iterations
    n_warmup = variable(int, value=100)

    # Define which executable to run
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable
    executable = 'python3'

    # Define which options to pass to the executable
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.executable_opts
    executable_opts = ['mpi4py_reduce.py', '--n_iter', f'{n_iterations}', '--n_warmup', f'{n_warmup}']

    # Temporarily define postrun_cmds to make it easy to find out memory usage
    postrun_cmds = [
        # for cgroups v1
        'MAX_MEM_IN_BYTES=$(</sys/fs/cgroup/memory/$(</proc/self/cpuset)/../memory.max_usage_in_bytes)',
        # for cgroups v2
        # 'MAX_MEM_IN_BYTES=$(</sys/fs/cgroup/$(</proc/self/cpuset)/../../../memory.peak)',
        'echo "MAX_MEM_IN_BYTES=$MAX_MEM_IN_BYTES"',
        'echo "MAX_MEM_IN_MIB=$(($MAX_MEM_IN_BYTES/1048576))"'
    ]

    # Define a time limit for the scheduler running this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.pipeline.RegressionTest.time_limit
    time_limit = '5m00s'

    @run_after('init')
    def set_modules(self):
        hooks.set_modules(self)

    # Using this decorator, we tell ReFrame to run this AFTER the init step of the test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.run_after
    # See https://reframe-hpc.readthedocs.io/en/stable/pipeline.html for all steps in the pipeline
    # that reframe uses to execute tests.
    @run_after('init')
    def run_after_init(self):
        hooks.set_tag_scale(self)

    @run_after('setup')
    def set_num_tasks_per_node(self):
        """ Setting number of tasks per node and cpus per task in this function. This function sets
        num_tasks, num_tasks_per_node, num_cpus_per_task, and num_gpus_per_node, based on the current scale
        and the current partition's num_cpus, max_avail_gpus_per_node and num_nodes"""
        hooks.assign_tasks_per_compute_unit(self, COMPUTE_UNIT[CPU])

        # This test scales almost indefinitely
        # For tests that have limited scaling, make sure that test instances exceeding
        # a predefined maximum task count are skipped using:
        # max_tasks = 300
        # self.skip_if(self.num_tasks > max_tasks,
        #              f'Skipping test: more than {max_tasks} tasks are requested ({self.num_tasks})')

    # Make sure we request sufficient memory from the scheduler
    @run_after('setup')
    def request_mem(self):
        mem_required = self.num_tasks_per_node * 256  # request 256 MB per task per node
        hooks.req_memory_per_node(self, app_mem_req=mem_required)

    # Set binding strategy
    @run_after('setup')
    def set_binding(self):
        hooks.set_compact_process_binding(self)

    # Now, we check if the pattern 'Sum of all ranks: X' with X the correct sum for the amount of ranks is found
    # in the standard output:
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.sanity_function
    @sanity_function
    def validate(self):
        # Sum of 0, ..., N-1 is (N * (N-1) / 2)
        sum_of_ranks = round(self.num_tasks * ((self.num_tasks - 1) / 2))
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.assert_found
        return sn.assert_found(r'Sum of all ranks: %s' % sum_of_ranks, self.stdout)

    # Now, we define a pattern to extract a number that reflects the performance of this test
    # https://reframe-hpc.readthedocs.io/en/stable/regression_test_api.html#reframe.core.builtins.performance_function
    @performance_function('s')
    def time(self):
        # https://reframe-hpc.readthedocs.io/en/stable/deferrable_functions_reference.html#reframe.utility.sanity.extractsingle
        return sn.extractsingle(r'^Time elapsed:\s+(?P<perf>\S+)', self.stdout, 'perf', float)

    @performance_function('MiB')
    def max_mem_in_mib(self):
        return sn.extractsingle(r'^MAX_MEM_IN_MIB=(?P<perf>\S+)', self.stdout, 'perf', int)



import setuptools
import sys
import pkg_resources


def get_version_by_import():
    # Add the fallback version to whatever was set for scm_dict
    sys.path.append('.')
    from eessi.testsuite import __version__
    return __version__


# Get python version
python_version = sys.version_info

# Get setuptools version
# We control it when installing from pyproject.toml, but not when installing from setup.py / setup.cfg
# Check setuptools version
current_setuptools_version = pkg_resources.parse_version(pkg_resources.get_distribution("setuptools").version)

# Set the version requirement for setuptools_scm, depending on the combination of Python and setuptools version
version_file_path = 'eessi/testsuite/_version.py'
scm_dict = {'write_to': version_file_path}
if python_version >= (3, 8) and current_setuptools_version >= pkg_resources.parse_version("61.0.0"):
    setuptools_scm_requirement = 'setuptools_scm>8.0.0,<=8.1.0'
    scm_dict = {'version_file': version_file_path}
elif python_version >= (3, 7) and current_setuptools_version >= pkg_resources.parse_version("45.0.0"):
    setuptools_scm_requirement = 'setuptools_scm>7.0.0,<=7.1.0'
elif python_version >= (3, 6) and current_setuptools_version >= pkg_resources.parse_version("45.0.0"):
    setuptools_scm_requirement = 'setuptools_scm>=6.0.0,<=6.4.2'
elif python_version >= (3, 6) and current_setuptools_version >= pkg_resources.parse_version("42.0.0"):
    setuptools_scm_requirement = 'setuptools_scm>=5.0.0,<=5.0.2'
elif python_version >= (3, 6) and current_setuptools_version >= pkg_resources.parse_version("34.4.0"):
    setuptools_scm_requirement = 'setuptools_scm>=4.0.0,<=4.1.2'

# Set the fallback_version for scm based on what eessi.testsuite.__version__ returns
scm_dict['fallback_version'] = get_version_by_import()

setuptools.setup(
    use_scm_version=scm_dict,
    setup_requires=[setuptools_scm_requirement],
)



from reframe.core.builtins import parameter, run_after, run_before, variable
from reframe.core.exceptions import ReframeFatalError
from reframe.core.pipeline import RegressionMixin
from reframe.utility.sanity import make_performance_function
import reframe.utility.sanity as sn

from eessi.testsuite import hooks
from eessi.testsuite.constants import CI, DEVICE_TYPES, SCALES, COMPUTE_UNIT, TAGS
from eessi.testsuite.utils import log
from eessi.testsuite import __version__ as testsuite_version


# Hooks from the Mixin class seem to be executed _before_ those of the child class
# Thus, if the Mixin class needs self.X to be defined in after setup, the child class would have to define it before
# setup. That's a disadvantage and might not always be possible - let's see how far we get. It also seems that,
# like normal inheritance, functions with the same in the child and parent class will mean the child class
# will overwrite that of the parent class. That is a plus, as we can use the EESSI_Mixin class as a basis,
# but still overwrite specific functions in case specific tests would require this
# TODO: for this reason, we probably want to put _each_ hooks.something invocation into a seperate function,
# so that each individual one can be overwritten
#
# Note that I don't think we can do anything about the things set in the class body, such as the parameter's.
# Maybe we can move those to an __init__ step of the Mixin, even though that is not typically how ReFrame
# does it anymore?
# That way, the child class could define it as class variables, and the parent can use it in its __init__ method?
class EESSI_Mixin(RegressionMixin):
    """
    All EESSI tests should derive from this mixin class unless they have a very good reason not to.
    To run correctly, tests inheriting from this class need to define variables and parameters that are used here.
    That definition needs to be done 'on time', i.e. early enough in the execution of the ReFrame pipeline.
    Here, we list which class attributes must be defined by the child class, and by (the end of) what phase:

    - Init phase: device_type, scale, module_name, bench_name (if bench_name_ci is set)
    - Setup phase: compute_unit, required_mem_per_node

    The child class may also overwrite the following attributes:

    - Init phase: time_limit, measure_memory_usage, bench_name_ci
    """

    # Set defaults for these class variables, can be overwritten by child class if desired
    measure_memory_usage = variable(bool, value=False)
    exact_memory = variable(bool, value=False)
    user_executable_opts = variable(str, value='')
    scale = parameter(SCALES.keys())
    bench_name = None
    bench_name_ci = None
    num_tasks_per_compute_unit = 1
    always_request_gpus = None

    # Create ReFrame variables for logging runtime environment information
    cvmfs_repo_name = variable(str, value='None')
    cvmfs_software_subdir = variable(str, value='None')
    full_modulepath = variable(str, value='None')
    # These are optionally set in CI on the command line
    EESSI_CONFIGS_URL = variable(str, value='None')
    EESSI_CONFIGS_BRANCH = variable(str, value='None')

    # Make sure the version of the EESSI test suite gets logged in the ReFrame report
    eessi_testsuite_version = variable(str, value=testsuite_version)

    # Note that the error for an empty parameter is a bit unclear for ReFrame 4.6.2, but that will hopefully improve
    # see https://github.com/reframe-hpc/reframe/issues/3254
    # If that improves: uncomment the following to force the user to set module_name
    # module_name = parameter()

    def __init_subclass__(cls, **kwargs):
        " set default values for built-in ReFrame attributes "
        super().__init_subclass__(**kwargs)
        cls.valid_prog_environs = ['default']
        cls.valid_systems = ['*']
        if not cls.time_limit:
            cls.time_limit = '1h'
        if not cls.readonly_files:
            msg = ' '.join([
                "Built-in attribute `readonly_files` is empty. To avoid excessive copying, it's highly recommended",
                "to add all files and/or dirs in `sourcesdir` that are needed but not modified during the test,",
                "thus can be symlinked into the stage dirs. If you are sure there are no such files,",
                "set `readonly_files = ['']`.",
            ])
            raise ReframeFatalError(msg)

    # Helper function to validate if an attribute is present it item_dict.
    # If not, print it's current name, value, and the valid_values
    def EESSI_mixin_validate_item_in_dict(self, item, item_dict, check_keys=False):
        """
        Check if the item 'item' exist in the values of 'item_dict'.
        If check_keys=True, then it will check instead of 'item' exists in the keys of 'item_dict'.
        If item is not found, an error will be raised that will mention the valid values for 'item'.
        """
        if check_keys:
            valid_items = list(item_dict.keys())
        else:
            valid_items = list(item_dict.values())

        value = getattr(self, item)
        if value not in valid_items:
            if len(valid_items) == 1:
                msg = f"The variable '{item}' has value {value}, but the only valid value is {valid_items[0]}"
            else:
                msg = f"The variable '{item}' has value {value}, but the only valid values are {valid_items}"
            raise ReframeFatalError(msg)

    @run_after('init')
    def EESSI_mixin_validate_init(self):
        """Check that all variables that have to be set for subsequent hooks in the init phase have been set"""
        # List which variables we will need/use in the run_after('init') hooks
        var_list = ['device_type', 'scale', 'module_name', 'measure_memory_usage']
        for var in var_list:
            if not hasattr(self, var):
                msg = "The variable '%s' should be defined in any test class that inherits" % var
                msg += " from EESSI_Mixin before (or in) the init phase, but it wasn't"
                raise ReframeFatalError(msg)

        # Check that the value for these variables is valid,
        # i.e. exists in their respective dict from eessi.testsuite.constants
        self.EESSI_mixin_validate_item_in_dict('device_type', DEVICE_TYPES)
        self.EESSI_mixin_validate_item_in_dict('scale', SCALES, check_keys=True)
        self.EESSI_mixin_validate_item_in_dict('valid_systems', {'valid_systems': ['*']})
        self.EESSI_mixin_validate_item_in_dict('valid_prog_environs', {'valid_prog_environs': ['default']})

    @run_after('init')
    def EESSI_mixin_run_after_init(self):
        """Hooks to run after init phase"""

        # Filter on which scales are supported by the partitions defined in the ReFrame configuration
        hooks.filter_supported_scales(self)

        hooks.filter_valid_systems_by_device_type(self, required_device_type=self.device_type)

        hooks.set_modules(self)

        # Set scales as tags
        hooks.set_tag_scale(self)

    @run_before('setup', always_last=True)
    def EESSI_mixin_measure_mem_usage(self):
        if self.measure_memory_usage:
            hooks.measure_memory_usage(self)
            # Since we want to do this conditionally on self.measure_mem_usage, we use make_performance_function
            # instead of the @performance_function decorator
            self.perf_variables['memory'] = make_performance_function(hooks.extract_memory_usage, 'MiB', self)

    @run_after('init', always_last=True)
    def EESSI_mixin_set_tag_ci(self):
        """
        Set CI tag if bench_name_ci and bench_name are set and are equal
        Also set tag on bench_name if set
        """
        tags_added = False
        if self.bench_name_ci:
            if not self.bench_name:
                msg = "Attribute bench_name_ci is set, but bench_name is not set"
                raise ReframeFatalError(msg)
            if self.bench_name == self.bench_name_ci:
                self.tags.add(TAGS[CI])
                tags_added = True
        if self.bench_name:
            self.tags.add(self.bench_name)
            tags_added = True
        if tags_added:
            log(f'tags set to {self.tags}')

    @run_after('setup')
    def EESSI_mixin_validate_setup(self):
        """Check that all variables that have to be set for subsequent hooks in the setup phase have been set"""
        var_list = ['compute_unit']
        for var in var_list:
            if not hasattr(self, var):
                msg = "The variable '%s' should be defined in any test class that inherits" % var
                msg += " from EESSI_Mixin before (or in) the setup phase, but it wasn't"
                raise ReframeFatalError(msg)

        # Check if mem_func was defined to compute the required memory per node as function of the number of
        # tasks per node
        if not hasattr(self, 'required_mem_per_node'):
            msg = "The function 'required_mem_per_node' should be defined in any test class that inherits"
            msg += " from EESSI_Mixin before (or in) the setup phase, but it wasn't. Note that this function"
            msg += " can use self.num_tasks_per_node, as it will be called after that attribute"
            msg += " has been set."
            raise ReframeFatalError(msg)

        # Check that the value for these variables is valid
        # i.e. exists in their respective dict from eessi.testsuite.constants
        self.EESSI_mixin_validate_item_in_dict('compute_unit', COMPUTE_UNIT)

    @run_after('setup')
    def EESSI_mixin_assign_tasks_per_compute_unit(self):
        """Call hooks to assign tasks per compute unit, set OMP_NUM_THREADS, and set compact process binding"""
        hooks.assign_tasks_per_compute_unit(test=self, compute_unit=self.compute_unit,
                                            num_per=self.num_tasks_per_compute_unit)

        # Set OMP_NUM_THREADS environment variable
        hooks.set_omp_num_threads(self)

        # Set compact process binding
        hooks.set_compact_process_binding(self)

    @run_after('setup')
    def EESSI_mixin_request_mem(self):
        """Call hook to request the required amount of memory per node"""
        hooks.req_memory_per_node(self, app_mem_req=self.required_mem_per_node())

    @run_after('setup')
    def EESSI_mixin_log_runtime_info(self):
        """Log additional runtime information: which CVMFS repo was used (or if it was testing local software),
        path to the modulefile, EESSI software subdir, EESSI testsuite version"""
        self.postrun_cmds.append('echo "EESSI_CVMFS_REPO: $EESSI_CVMFS_REPO"')
        self.postrun_cmds.append('echo "EESSI_SOFTWARE_SUBDIR: $EESSI_SOFTWARE_SUBDIR"')
        if self.module_name:
            # Get full modulepath
            get_full_modpath = f'echo "FULL_MODULEPATH: $(module --location show {self.module_name})"'
            self.postrun_cmds.append(get_full_modpath)

    @run_before('run', always_last=True)
    def EESSI_mixin_set_user_executable_opts(self):
        "Override executable_opts with user_executable_opts if set on the cmd line"
        if self.user_executable_opts:
            log(f'Overwriting executable_opts {self.executable_opts} by executable_opts '
                'specified on cmd line {[self.user_executable_opts]}')
            self.executable_opts = [self.user_executable_opts]

    @run_after('run')
    def EESSI_mixin_extract_runtime_info_from_log(self):
        """Extracts the printed runtime info from the job log and logs it as reframe variables"""
        if self.is_dry_run():
            return

        # If EESSI_CVMFS_REPO environment variable was set, extract it and store it in self.cvmfs_repo_name
        repo_name = sn.extractall(r'EESSI_CVMFS_REPO: /cvmfs/(?P<repo>.*)$', f'{self.stagedir}/{self.stdout}',
                                  'repo', str)
        if repo_name:
            self.cvmfs_repo_name = f'{repo_name}'

        software_subdir = sn.extractall(r'EESSI_SOFTWARE_SUBDIR: (?P<subdir>.*)$',
                                        f'{self.stagedir}/{self.stdout}', 'subdir', str)
        if software_subdir:
            self.cvmfs_software_subdir = f'{software_subdir}'

        module_path = sn.extractall(r'FULL_MODULEPATH: (?P<modpath>.*)$', f'{self.stagedir}/{self.stdout}',
                                    'modpath', str)
        if module_path:
            self.full_modulepath = f'{module_path}'



# WARNING: for CPU autodetect to work correctly you need to
# 1. Either use ReFrame >= 4.3.3 or temporarily change the 'launcher' for each partition to srun
# 2. Either use ReFrame >= 4.3.3 or run from a clone of the ReFrame repository
# 3. Temporarily change the 'access' field for the GPU partition to
# 'access':  ['-p gpu', '--export=None', '--exclusive'],

# Without this, the autodetect job fails because
# 1. A missing mpirun command
# 2. An incorrect directory structure is assumed when preparing the stagedir for the autodetect job
# 3. Snellius doesn't allow submission to the GPU partition without requesting at least one GPU

# Related issues
# 1. https://github.com/reframe-hpc/reframe/issues/2926
# 2. https://github.com/reframe-hpc/reframe/issues/2914

import os

from eessi.testsuite.common_config import common_logging_config, common_general_config, common_eessi_init
from eessi.testsuite.constants import *  # noqa: F403

# This config will write all staging, output and logging to subdirs under this prefix
# Override with RFM_PREFIX environment variable
reframe_prefix = os.path.join(os.environ['HOME'], 'reframe_runs')

# Jobs that partially fill multiple nodes are not allowed on the GPU partition
valid_scales_snellius_gpu = [s for s in SCALES if s not in ['1cpn_2nodes', '1cpn_4nodes']]

# This is an example configuration file
site_configuration = {
    'systems': [
        {
            'name': 'snellius',
            'descr': 'Dutch National Supercomputer',
            'modules_system': 'lmod',
            # Just accept any hostname. List of hostnames can only get outdated, and this is a single-system config
            # file anyway.
            'hostnames': ['.*'],
            'prefix': reframe_prefix,
            'stagedir': f'/scratch-shared/{os.environ.get("USER")}/reframe_output/staging',
            'partitions': [
                {
                    'name': 'rome',
                    'scheduler': 'slurm',
                    'prepare_cmds': [common_eessi_init()],
                    'launcher': 'mpirun',
                    'access': ['-p rome', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 120,
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 229376  # in MiB
                    },
                    'descr': 'AMD Rome CPU partition with native EESSI stack'
                },
                {
                    'name': 'genoa',
                    'scheduler': 'slurm',
                    'prepare_cmds': [
                        # EESSI init script (for now) falls back to zen3, since the zen4 is incomplete
                        # But, we want to really test the zen4 branch on these nodes
                        'export EESSI_SOFTWARE_SUBDIR_OVERRIDE=x86_64/amd/zen4',
                        common_eessi_init()
                    ],
                    'launcher': 'mpirun',
                    'access': ['-p genoa', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 120,
                    'resources': [
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[CPU],
                    ] + list(SCALES.keys()),
                    'extras': {
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 344064  # in MiB
                    },
                    'descr': 'AMD Genoa CPU partition with native EESSI stack'
                },
                {
                    'name': 'gpu_A100',
                    'scheduler': 'slurm',
                    'prepare_cmds': [common_eessi_init()],
                    'launcher': 'mpirun',
                    'access': ['-p gpu_a100', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 60,
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[GPU],
                        FEATURES[ALWAYS_REQUEST_GPUS],
                    ] + valid_scales_snellius_gpu,
                    'extras': {
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 491520  # in MiB
                    },
                    'descr': 'Nvidia A100 GPU partition with native EESSI stack'
                },
                {
                    'name': 'gpu_H100',
                    'scheduler': 'slurm',
                    'prepare_cmds': [common_eessi_init()],
                    'launcher': 'mpirun',
                    'access': ['-p gpu_h100', '--export=None'],
                    'environs': ['default'],
                    'max_jobs': 60,
                    'devices': [
                        {
                            'type': DEVICE_TYPES[GPU],
                            'num_devices': 4,
                        }
                    ],
                    'resources': [
                        {
                            'name': '_rfm_gpu',
                            'options': ['--gpus-per-node={num_gpus_per_node}'],
                        },
                        {
                            'name': 'memory',
                            'options': ['--mem={size}'],
                        }
                    ],
                    'features': [
                        FEATURES[GPU],
                        FEATURES[ALWAYS_REQUEST_GPUS],
                    ] + valid_scales_snellius_gpu,
                    'extras': {
                        GPU_VENDOR: GPU_VENDORS[NVIDIA],
                        # Make sure to round down, otherwise a job might ask for more mem than is available
                        # per node
                        'mem_per_node': 737280  # in MiB
                    },
                    'descr': 'Nvidia H100 GPU partition with native EESSI stack'
                },

            ]
        },
    ],
    'environments': [
        {
            'name': 'default',
            'cc': 'cc',
            'cxx': '',
            'ftn': '',
        },
    ],
    'logging': common_logging_config(reframe_prefix),
    'general': [
        {
            # Enable automatic detection of CPU architecture for each partition
            # See https://reframe-hpc.readthedocs.io/en/stable/configure.html#auto-detecting-processor-information
            'remote_detect': True,
            **common_general_config(reframe_prefix)
        }
    ],
}



import argparse
import timeit
import os

import numpy as np

import torch.backends.cudnn as cudnn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data.distributed
from torchvision import models


# Benchmark settings
parser = argparse.ArgumentParser(description='PyTorch Synthetic Benchmark',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--fp16-allreduce', action='store_true', default=False,
                    help='use fp16 compression during allreduce')

parser.add_argument('--model', type=str, default='resnet50',
                    help='model to benchmark')
parser.add_argument('--batch-size', type=int, default=32,
                    help='input batch size')

parser.add_argument('--num-warmup-batches', type=int, default=10,
                    help='number of warm-up batches that don\'t count towards benchmark')
parser.add_argument('--num-batches-per-iter', type=int, default=10,
                    help='number of batches per benchmark iteration')
parser.add_argument('--num-iters', type=int, default=10,
                    help='number of benchmark iterations')

parser.add_argument('--no-cuda', action='store_true', default=False,
                    help='disables CUDA training')

parser.add_argument('--use-adasum', action='store_true', default=False,
                    help='use adasum algorithm to do reduction')
parser.add_argument('--use-horovod', action='store_true', default=False)
parser.add_argument('--use-ddp', action='store_true', default=False)

parser.add_argument('--use-amp', action='store_true', default=False,
                    help='Use PyTorch Automatic Mixed Precision (AMP)')
parser.add_argument('--world-size', type=int, default=1,
                    help='Define the world size for ddp')
parser.add_argument('--master-port', type=int, default=False,
                    help='Define a master port for ddp')
parser.add_argument('--master-address', type=str, default='localhost',
                    help='Define a master address for ddp')

args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if args.use_horovod and args.use_ddp:
    print("You can't specify to use both Horovod and Pytorch DDP, exiting...")
    exit(1)

# Set MASTER_ADDR and MASTER_PORT environment variables
# By doing it as part of this python script, we don't need to have the launchers export them
# This saves us from having to find a launcher-agnostic way of exporting variables
os.environ['MASTER_ADDR'] = args.master_address
os.environ['MASTER_PORT'] = '%s' % args.master_port

# Set a default rank and world size, also for when ddp and horovod are not used
rank = 0
world_size = args.world_size
if args.use_horovod:
    import horovod.torch as hvd
    hvd.init()
    rank = hvd.local_rank()
    world_size = hvd.size()

    if args.cuda:
        # If launched with srun, you are in a CGROUP with only 1 GPU, so you don't need to set it.
        # If launched with mpirun, you see ALL local GPUs on the node, and you need to set which one
        # this rank should use.
        visible_gpus = torch.cuda.device_count()
        # Horovod: pin GPU to local rank.
        if visible_gpus > 1:
            torch.cuda.set_device(hvd.local_rank())

    # Should only be uncommented for debugging
    # In ReFrame tests, a print from each rank can mess up the output file, causing
    # performance and sanity patterns to not be found
    # print(f"hvd.local_rank: {rank}", flush=True)


if args.use_ddp:
    from socket import gethostname
    import torch.distributed as dist
    from torch.nn.parallel import DistributedDataParallel as DDP

    # world_size = int(os.environ["SLURM_NTASKS"])  ## No longer needed now we pass it as argument?
    # If launched with mpirun, get rank from this
    rank = int(os.environ.get("OMPI_COMM_WORLD_RANK", -1))
    if rank == -1:
        # Else it's launched with srun, get rank from this
        rank = int(os.environ.get("SLURM_PROCID", -1))
    if rank == -1:
        err_msg = "ERROR: cannot determine local rank. This test currently only supports OpenMPI"
        err_msg += " and srun as launchers. If you've configured a different launcher for your system"
        err_msg += " this test will need to be extended with a method to get it's local rank for that launcher."
        print(err_msg)

    # If launched with srun, you are in a CGROUP with only 1 GPU, so you don't need to set it.
    # If launched with mpirun, you see ALL local GPUs on the node, and you need to set which one
    # this rank should use.
    visible_gpus = torch.cuda.device_count()
    if visible_gpus > 1:
        print("Listing visible devices")
        for i in range(torch.cuda.device_count()):
            print(f"Device {i}: {torch.cuda.device(i)}")
        # This assumes compact mapping of ranks to available hardware
        # e.g. rank 0-x to node 1, rank x-y to node 2, etc
        # Assuming the set_compact_process_binding hook from the EESSI testsuite is called,
        # this condition should be satisfied
        local_rank = rank - visible_gpus * (rank // visible_gpus)
        torch.cuda.set_device(local_rank)
        print("Listing visible devices after setting one")
        for i in range(torch.cuda.device_count()):
            print(f"Device {i}: {torch.cuda.device(i)}")
        # We should also set CUDA_VISIBLE_DEVICES, which gets respected by NCCL
        os.environ['CUDA_VISIBLE_DEVICES'] = '%s' % local_rank
        print(f"host: {gethostname()}, rank: {rank}, local_rank: {local_rank}")
    else:
        print(f"host: {gethostname()}, rank: {rank}")

    def setup(rank, world_size):

        # initialize the process group
        if args.cuda:
            dist.init_process_group("nccl", rank=rank, world_size=world_size)
        else:
            dist.init_process_group("gloo", rank=rank, world_size=world_size)

    def cleanup():
        # clean up the distributed environment
        dist.destroy_process_group()

    setup(rank, world_size)
    if rank == 0:
        print(f"Group initialized? {dist.is_initialized()}", flush=True)


# This relies on the 'rank' set in the if args.use_horovod or args.use_ddp sections
def log(s, nl=True):
    if (args.use_horovod or args.use_ddp) and rank != 0:
        return
    print(s, end='\n' if nl else '', flush=True)


log(f"World size: {world_size}")

torch.set_num_threads(int(os.environ['OMP_NUM_THREADS']))
torch.set_num_interop_threads(2)

cudnn.benchmark = True

# Set up standard model.
model = getattr(models, args.model)()

# By default, Adasum doesn't need scaling up learning rate.
lr_scaler = hvd.size() if not args.use_adasum and args.use_horovod else 1

if args.cuda:
    # Move model to GPU.
    model.cuda()
    # If using GPU Adasum allreduce, scale learning rate by local_size.
    if args.use_horovod and args.use_adasum and hvd.nccl_built():
        lr_scaler = hvd.local_size()

# If using DDP, wrap model
if args.use_ddp:
    model = DDP(model)

optimizer = optim.SGD(model.parameters(), lr=0.01 * lr_scaler)

# Horovod: (optional) compression algorithm.
if args.use_horovod:
    compression = hvd.Compression.fp16 if args.fp16_allreduce else hvd.Compression.none

# Horovod: wrap optimizer with DistributedOptimizer.
if args.use_horovod:
    optimizer = hvd.DistributedOptimizer(optimizer,
                                         named_parameters=model.named_parameters(),
                                         compression=compression,
                                         op=hvd.Adasum if args.use_adasum else hvd.Average)

    # Horovod: broadcast parameters & optimizer state.
    hvd.broadcast_parameters(model.state_dict(), root_rank=0)
    hvd.broadcast_optimizer_state(optimizer, root_rank=0)

# Set up fixed fake data
data = torch.randn(args.batch_size, 3, 224, 224)
target = torch.LongTensor(args.batch_size).random_() % 1000
if args.cuda:
    data, target = data.cuda(), target.cuda()

# Create GradScaler for automatic mixed precision
scaler = torch.cuda.amp.GradScaler(enabled=args.use_amp)

# Set device_type for AMP
if args.cuda:
    device_type = "cuda"
else:
    device_type = "cpu"


def benchmark_step():
    optimizer.zero_grad()
    with torch.autocast(device_type=device_type, enabled=args.use_amp):
        output = model(data)
        loss = F.cross_entropy(output, target)
    scaler.scale(loss).backward()
    scaler.step(optimizer)
    scaler.update()


log('Model: %s' % args.model)
log('Batch size: %d' % args.batch_size)
device = 'GPU' if args.cuda else 'CPU'
if args.use_horovod:
    log('Number of %ss: %d' % (device, hvd.size()))

# Warm-up
log('Running warmup...')
timeit.timeit(benchmark_step, number=args.num_warmup_batches)

# Benchmark
log('Running benchmark...')
img_secs = []
for x in range(args.num_iters):
    time = timeit.timeit(benchmark_step, number=args.num_batches_per_iter)
    img_sec = args.batch_size * args.num_batches_per_iter / time
    log('Iter #%d: %.1f img/sec per %s' % (x, img_sec, device))
    img_secs.append(img_sec)

# Results
img_sec_mean = np.mean(img_secs)
img_sec_conf = 1.96 * np.std(img_secs)
log('Img/sec per %s: %.1f +-%.1f' % (device, img_sec_mean, img_sec_conf))
log('Total img/sec on %d %s(s): %.1f +-%.1f' %
    (world_size, device, world_size * img_sec_mean, world_size * img_sec_conf))



"""
Constants for ReFrame tests
"""

AMD = 'AMD'
CI = 'CI'
HWTHREAD = 'HWTHREAD'
CPU = 'CPU'
CPU_SOCKET = 'CPU_SOCKET'
NUMA_NODE = 'NUMA_NODE'
GPU = 'GPU'
GPU_VENDOR = 'GPU_VENDOR'
INTEL = 'INTEL'
NODE = 'NODE'
NVIDIA = 'NVIDIA'
ALWAYS_REQUEST_GPUS = 'ALWAYS_REQUEST_GPUS'

DEVICE_TYPES = {
    CPU: 'cpu',
    GPU: 'gpu',
}

COMPUTE_UNIT = {
    HWTHREAD: 'hwthread',
    CPU: 'cpu',
    CPU_SOCKET: 'cpu_socket',
    NUMA_NODE: 'numa_node',
    GPU: 'gpu',
    NODE: 'node',
}

TAGS = {
    CI: 'CI',
}

FEATURES = {
    CPU: 'cpu',
    GPU: 'gpu',
    ALWAYS_REQUEST_GPUS: 'always_request_gpus',
}

GPU_VENDORS = {
    AMD: 'amd',
    INTEL: 'intel',
    NVIDIA: 'nvidia',
}

SCALES = {
    # required keys:
    # - num_nodes
    # - either node_part or (num_cpus_per_node and num_gpus_per_node)
    # num_cpus_per_node and num_gpus_per_node are upper limits:
    # the actual count depends on the specific configuration of cores, gpus, and sockets within the node,
    # as well as the specific test being carried out.
    '1_core': {'num_nodes': 1, 'num_cpus_per_node': 1, 'num_gpus_per_node': 1},
    '2_cores': {'num_nodes': 1, 'num_cpus_per_node': 2, 'num_gpus_per_node': 1},
    '4_cores': {'num_nodes': 1, 'num_cpus_per_node': 4, 'num_gpus_per_node': 1},
    # renamed after v0.2.0 from 1_cpn_2_nodes to make more unique
    '1cpn_2nodes': {'num_nodes': 2, 'num_cpus_per_node': 1, 'num_gpus_per_node': 1},
    # renamed after v0.2.0 from 1_cpn_4_nodes to make more unique
    '1cpn_4nodes': {'num_nodes': 4, 'num_cpus_per_node': 1, 'num_gpus_per_node': 1},
    '1_8_node': {'num_nodes': 1, 'node_part': 8},  # 1/8 node
    '1_4_node': {'num_nodes': 1, 'node_part': 4},  # 1/4 node
    '1_2_node': {'num_nodes': 1, 'node_part': 2},  # 1/2 node
    '1_node': {'num_nodes': 1, 'node_part': 1},
    '2_nodes': {'num_nodes': 2, 'node_part': 1},
    '4_nodes': {'num_nodes': 4, 'node_part': 1},
    '8_nodes': {'num_nodes': 8, 'node_part': 1},
    '16_nodes': {'num_nodes': 16, 'node_part': 1},
}

# When tests are filtered by the hooks, the valid_systems is set to this system name:
INVALID_SYSTEM = "INVALID_SYSTEM"
